package demo;

import java.sql.*;
import java.util.*;
import java.net.URLDecoder;
import java.io.UnsupportedEncodingException;

import spark.Request;
import spark.Response;

import com.google.gson.*;

/**
   ,* Database is an interface to the college application database, it
   ,* uses JDBC to connect to a SQLite3 file.
   ,*/
public class Database {

    /**
       ,* The database connection.
       ,*/
    private Connection conn;
    private Gson gson = new Gson();

    /**
     * Creates the database interface object. Connection to the
     * database is performed later.
     */
    public Database() {
        conn = null;
    }

    /**
     * Opens a connection to the database, using the specified
     * filename (if we'd used a traditional DBMS, such as PostgreSQL
     * or MariaDB, we would have specified username and password
     * instead).
     */
    public boolean openConnection(String filename) {
        try {
            conn = DriverManager.getConnection("jdbc:sqlite:" + filename);
        } catch (SQLException e) {
            e.printStackTrace();
            return false;
        } catch (Exception e) {
            e.printStackTrace();
            return false;
        }
        return true;
    }

    /**
     * Closes the connection to the database.
     */
    public void closeConnection() {
        try {
            if (conn != null) {
                conn.close();
            }
        } catch (SQLException e) {
            e.printStackTrace();
        }
    }

    /**
     * Checks if the connection to the database has been established
     * 
     * @return true if the connection has been established
     */
    public boolean isConnected() {
        return conn != null;
    }

    /* ================================== */
    /* --- insert your own code below --- */
    /* ===============================*== */

    class DataWrapper {
        private Object data;

        public DataWrapper  (Object data) {
            this.data = data;
        }
    }

    private String asJsonData(Object object) {
        return gson.toJson(new DataWrapper(object));
    }

    public String getStudent(Request req,
                             Response res,
                             int studentId) {
        var found = new ArrayList<Student>();
        var query =
            """
            SELECT    s_id, s_name, gpa, size_hs
            FROM      students
            WHERE     s_id = ?
            """;
        try (var ps = conn.prepareStatement(query)) {
            ps.setInt(1, studentId);
            var rs = ps.executeQuery();
            while (rs.next()) {
                found.add(Student.fromRS(rs));
            }
        } catch (SQLException e) {
            e.printStackTrace();
        }
        res.status(200);
        return asJsonData(found);
    }

    private String checkParam(Request req,
                              String paramName,
                              String condition,
                              List<String> params) {
        var param = req.queryParams(paramName);
        if (param != null) {
            try {
                params.add(URLDecoder.decode(param, "utf-8"));
                return '\n' + condition;   // put each condition on its own line
            } catch (UnsupportedEncodingException e) {
                System.err.printf("Weird encoding: %s\n", e);
            }
        }
        return "";   // nothing to add, since the parameter was absent
    }

    public String getStudents(Request req, Response res) {
        var found = new ArrayList<Student>();
        var query =
            """
            SELECT    s_id, s_name, gpa, size_hs
            FROM      students
            WHERE     1 = 1
            """;
        var params = new ArrayList<String>();
        query += checkParam(req, "name", "AND s_name = ?", params);
        query += checkParam(req, "minGpa", "AND gpa >= ?", params);
        try (var ps = conn.prepareStatement(query)) {
            var idx = 0;
            for (var param : params) {
                ps.setString(++idx, param);
            }
            var rs = ps.executeQuery();
            while (rs.next()) {
                found.add(Student.fromRS(rs));
            }
        } catch (SQLException e) {
            e.printStackTrace();
        }
        res.status(200);
        return asJsonData(found);
    }

    public String postStudent(Request req, Response res) {
        var body = req.body();
        var student = Student.fromJson(gson, body);
        var statement =
            """
            INSERT
            INTO    students(s_id, s_name, gpa, size_hs)
            VALUES  (?,?,?,?)
            """;
        try (var ps = conn.prepareStatement(statement)) {
            ps.setInt(1, student.id);
            ps.setString(2, student.name);
            ps.setDouble(3, student.gpa);
            ps.setInt(4, student.sizeHS);
            ps.executeUpdate();
            res.status(201);
        } catch (SQLException e) {
            e.printStackTrace();
            res.status(400);
        }
        res.status(201);
        return String.format("http://localhost/students/%d", student.id);
    }

    public String getStudentApplications(Request req,
                                         Response res,
                                         int studentId) {
        var found = new ArrayList<Application>();
        var query =
            """
            SELECT   s_id, c_name, major, decision
            FROM     applications
            JOIN     students
            USING    (s_id)
            WHERE    s_id = ?
            """;
        try (var ps = conn.prepareStatement(query)) {
            ps.setInt(1, studentId);
            var rs = ps.executeQuery();
            while (rs.next()) {
                found.add(Application.fromRS(rs));
            }
        } catch (SQLException e) {
            e.printStackTrace();
        }
        res.status(200);
        return asJsonData(found);
    }
}

class Student {

    public final int id;
    public final String name;
    public final double gpa;
    public final int sizeHS;

    private Student  (int id, String name, double gpa, int sizeHS) {
        this.id = id;
        this.name = name;
        this.gpa = gpa;
        this.sizeHS = sizeHS;
    }

    public static Student fromRS(ResultSet rs) throws SQLException {
        return new Student(rs.getInt("s_id"),
                           rs.getString("s_name"),
                           rs.getDouble("gpa"),
                           rs.getInt("size_hs"));
    }

    public static Student fromJson(Gson gson, String json) {
        return gson.fromJson(json, Student.class);
    }

    public String toJson(Gson gson) {
        return gson.toJson(this);
    }
}

class Application {

    public final int id;
    public final String college;
    public final String major;
    public final boolean decision;

    private Application  (int id, String college, String major, boolean decision) {
        this.id = id;
        this.college = college;
        this.major = major;
        this.decision = decision;
    }

    public static Application fromRS(ResultSet rs) throws SQLException {
        return new Application(rs.getInt("s_id"),
                               rs.getString("c_name"),
                               rs.getString("major"),
                               rs.getString("decision").equals("Y"));
    }
}
